# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.utils import date as date_util
from orquesta.utils import jsonify as json_util


MOCK_DATETIME_STR = "2015-01-01T12:00:01.000000+01:00"

MOCK_JSON = {
    "k1": "abc",
    "k2": 123,
    "k3": False,
    "k4": MOCK_DATETIME_STR,
    "k5": [1, 3, 5, 7, 9],
    "k6": {"a": 1, "b": 2, "c": 3},
}

MOCK_JSON_UNSERIALIZEABLE = {"k1": object()}


class FakeModel(object):
    def __init__(self, *args, **kwargs):
        self.k1 = None
        self.k2 = None
        self.k3 = None
        self.k4 = None
        self.k5 = None
        self.k6 = None


class SerializationTest(unittest.TestCase):
    def test_serialize(self):
        obj = FakeModel()
        obj.k1 = MOCK_JSON["k1"]
        obj.k2 = MOCK_JSON["k2"]
        obj.k3 = MOCK_JSON["k3"]
        obj.k4 = MOCK_JSON["k4"]
        obj.k5 = MOCK_JSON["k5"]
        obj.k6 = MOCK_JSON["k6"]

        doc = json_util.serialize(obj)

        self.assertDictEqual(MOCK_JSON, doc)

    def test_serialize_unsupported_type(self):
        obj = FakeModel()
        obj.k1 = MOCK_JSON_UNSERIALIZEABLE["k1"]

        doc = json_util.serialize(obj)

        self.assertDictEqual(dict(), doc)

    def test_deserialize(self):
        obj = json_util.deserialize(FakeModel, MOCK_JSON)

        self.assertEqual(MOCK_JSON["k1"], obj.k1)
        self.assertEqual(MOCK_JSON["k2"], obj.k2)
        self.assertEqual(MOCK_JSON["k3"], obj.k3)
        self.assertEqual(date_util.parse(MOCK_JSON["k4"]), obj.k4)
        self.assertListEqual(MOCK_JSON["k5"], obj.k5)
        self.assertDictEqual(MOCK_JSON["k6"], obj.k6)

    def test_deserialize_unsupported_type(self):
        obj = json_util.deserialize(FakeModel, MOCK_JSON_UNSERIALIZEABLE)

        self.assertIsNone(obj.k1)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from orquesta import constants
from orquesta import exceptions as exc
from orquesta import statuses


def _get_current_task(context):
    if not context:
        raise exc.ExpressionEvaluationException("The context is not set.")

    try:
        current_task = context["__current_task"] or {}
    except KeyError:
        current_task = {}

    if not current_task:
        raise exc.ExpressionEvaluationException("The current task is not set in the context.")

    return current_task


def task_status_(context, task_id, route=None):
    if not context:
        return statuses.UNSET

    if route is None:
        try:
            current_task = context["__current_task"] or {}
            route = current_task["route"]
        except KeyError:
            route = 0

    try:
        workflow_state = context["__state"] or {}
    except KeyError:
        workflow_state = {}

    task_state_pointers = workflow_state.get("tasks") or {}
    task_state_entry_uid = constants.TASK_STATE_ROUTE_FORMAT % (task_id, str(route))
    task_state_entry_idx = task_state_pointers.get(task_state_entry_uid)

    # If unable to identify the task flow entry and if there are other routes, then
    # use an earlier route before the split to find the specific task.
    if task_state_entry_idx is None:
        if route > 0:
            current_route_details = workflow_state["routes"][route]
            # Reverse the list because we want to start with the next longest route.
            for idx, prev_route_details in enumerate(reversed(workflow_state["routes"][:route])):
                if len(set(prev_route_details) - set(current_route_details)) == 0:
                    # The index is from a reversed list so need to calculate
                    # the index of the item in the list before the reverse.
                    prev_route = route - idx - 1
                    return task_status_(context, task_id, route=prev_route)

        return statuses.UNSET

    task_seqs = workflow_state.get("sequence") or []
    task_state_entry = task_seqs[task_state_entry_idx]

    if task_state_entry is None:
        return statuses.UNSET

    return task_state_entry.get("status", statuses.UNSET)


def succeeded_(context):
    current_task = _get_current_task(context)
    task_id = current_task["id"]
    route = current_task["route"]

    return task_status_(context, task_id, route=route) == statuses.SUCCEEDED


def failed_(context):
    current_task = _get_current_task(context)
    task_id = current_task["id"]
    route = current_task["route"]

    return task_status_(context, task_id, route=route) == statuses.FAILED


def completed_(context):
    current_task = _get_current_task(context)
    task_id = current_task["id"]
    route = current_task["route"]

    return task_status_(context, task_id, route=route) in statuses.COMPLETED_STATUSES


def result_(context):
    current_task = _get_current_task(context)

    return current_task.get("result")


def item_(context, key=None):
    if not context:
        raise exc.ExpressionEvaluationException("The context is not set.")

    current_item = context["__current_item"]

    if not key:
        return current_item

    if not isinstance(current_item, collections.Mapping):
        raise exc.ExpressionEvaluationException("Item is not type of collections.Mapping.")

    if key not in current_item:
        raise exc.ExpressionEvaluationException('Item does not have key "%s".' % key)

    return current_item[key]



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import string

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorStressTest(test_base.WorkflowConductorTest):
    def _prep_wf_def(self, num_tasks):
        wf_def = {"input": ["data"], "tasks": {}}

        for i in range(1, num_tasks):
            task_name = "t" + str(i)
            next_task_name = "t" + str(i + 1)
            wf_def["tasks"][task_name] = {"action": "core.noop", "next": [{"do": next_task_name}]}

        task_name = "t" + str(num_tasks)
        wf_def["tasks"][task_name] = {"action": "core.noop"}

        return wf_def

    def _prep_conductor(self, num_tasks, context=None, inputs=None, status=None):
        wf_def = self._prep_wf_def(num_tasks)
        spec = native_specs.WorkflowSpec(wf_def)

        kwargs = {
            "context": context if context is not None else None,
            "inputs": inputs if inputs is not None else None,
        }

        conductor = conducting.WorkflowConductor(spec, **kwargs)

        if status:
            conductor.request_workflow_status(status)

        return conductor

    def test_runtime_function_of_graph_size(self):
        num_tasks = 100

        conductor = self._prep_conductor(num_tasks, status=statuses.RUNNING)

        for i in range(1, num_tasks + 1):
            task_name = "t" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_serialization_function_of_graph_size(self):
        num_tasks = 100
        conductor = self._prep_conductor(num_tasks, status=statuses.RUNNING)
        conductor.deserialize(conductor.serialize())

    def test_serialization_function_of_data_size(self):
        data_length = 1000000
        data = "".join(random.choice(string.ascii_lowercase) for _ in range(data_length))
        self.assertEqual(len(data), data_length)
        conductor = self._prep_conductor(1, inputs={"data": data}, status=statuses.RUNNING)
        conductor.deserialize(conductor.serialize())


class WorkflowConductorWithItemsStressTest(test_base.WorkflowConductorWithItemsTest):
    def test_runtime_function_of_items_list_size(self):
        wf_def = """
        version: 1.0

        vars:
          - xs: <% range(500).select(str($)) %>

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        num_items = 500

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": [str(i) for i in range(0, num_items)]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": i}, "item_id": int(i)}
            for i in task_ctx["xs"]
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED] * num_items
        expected_task_statuses = [statuses.RUNNING] * (num_items - 1) + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * (num_items - 1) + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert the workflow output is correct.
        conductor.render_workflow_output()
        expected_output = {"items": task_ctx["xs"]}
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import yaml


LOG = logging.getLogger(__name__)

FIXTURE_TYPES = ["workflows"]

FIXTURE_EXTS = {".json": json.load, ".yml": yaml.safe_load, ".yaml": yaml.safe_load}


def get_fixtures_base_path():
    return os.path.dirname(__file__)


def get_workflow_fixtures_base_path():
    return os.path.join(get_fixtures_base_path(), "workflows")


def get_fixture_content(fixture_file_name, fixture_type, raw=False):
    if fixture_type not in FIXTURE_TYPES:
        raise Exception("Unsupported fixture type of %s" % fixture_type)

    fixture_type_path = os.path.join(get_fixtures_base_path(), fixture_type)
    file_path = os.path.join(fixture_type_path, fixture_file_name)
    file_name, file_ext = os.path.splitext(file_path)

    if file_ext not in FIXTURE_EXTS.keys():
        raise Exception("Unsupported fixture file ext type of %s." % file_ext)

    with open(file_path, "r") as fd:
        return FIXTURE_EXTS[file_ext](fd) if not raw else fd.read()



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import statuses
from orquesta.tests.unit.conducting.native import base


class CyclicWorkflowConductorTest(base.OrchestraWorkflowConductorTest):
    def test_cycle(self):
        wf_name = "cycle"

        expected_task_seq = [
            "prep",
            "task1",
            "task2",
            "task3",
            "task1",
            "task2",
            "task3",
            "task1",
            "task2",
            "task3",
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq)

    def test_cycles(self):
        wf_name = "cycles"

        expected_task_seq = [
            "prep",
            "task1",
            "task2",
            "task3",
            "task4",
            "task2",
            "task5",
            "task1",
            "task2",
            "task3",
            "task4",
            "task2",
            "task5",
            "task1",
            "task2",
            "task3",
            "task4",
            "task2",
            "task5",
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq)

    def test_rollback_retry(self):
        wf_name = "rollback-retry"

        expected_task_seq = ["init", "check", "create", "rollback", "check", "delete"]

        mock_statuses = [
            statuses.SUCCEEDED,  # init
            statuses.FAILED,  # check
            statuses.SUCCEEDED,  # create
            statuses.SUCCEEDED,  # rollback
            statuses.SUCCEEDED,  # check
            statuses.SUCCEEDED,  # delete
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

    def test_cycle_and_fork(self):
        wf_name = "cycle-fork"

        expected_task_seq = [
            "init",
            "query",
            "decide_cheer",
            "decide_work",
            "cheer",
            "notify_work",
            "toil",
            "query",
            "decide_cheer",
            "decide_work",
        ]

        mock_results = [
            None,  # init
            True,  # query
            None,  # decide_cheer
            None,  # decide_work
            None,  # cheer
            None,  # notify_work
            None,  # toil
            False,  # query
            None,  # decide_cheer
            None,  # decide_work
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_results=mock_results)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six


def format_error(expr_type, expr, exc, spec_path=None, schema_path=None):
    error = {
        "type": expr_type,
        "expression": expr,
        "message": exc if isinstance(exc, six.string_types) else str(getattr(exc, "message", exc)),
    }

    if spec_path:
        error["spec_path"] = spec_path

    if schema_path:
        error["schema_path"] = schema_path

    return error



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.specs import loader as spec_loader
from orquesta.specs import mock as mock_specs


class SpecLoaderTest(unittest.TestCase):
    def test_get_bad_module(self):
        self.assertRaises(ImportError, spec_loader.get_spec_module, "foobar")

    def test_get_module(self):
        self.assertEqual(spec_loader.get_spec_module("mock"), mock_specs)

    def test_get_spec(self):
        spec_module = spec_loader.get_spec_module("mock")

        self.assertEqual(spec_module.WorkflowSpec.get_catalog(), "mock")
        self.assertEqual(spec_module.WorkflowSpec, mock_specs.WorkflowSpec)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from orquesta import exceptions as exc
from orquesta import statuses


LOG = logging.getLogger(__name__)


# Events for workflow execution.
WORKFLOW_REQUESTED = "workflow_requested"
WORKFLOW_SCHEDULED = "workflow_scheduled"
WORKFLOW_DELAYED = "workflow_delayed"
WORKFLOW_RUNNING = "workflow_running"
WORKFLOW_RUNNING_WORKFLOW_COMPLETED = "workflow_running_workflow_completed"
WORKFLOW_PAUSING = "workflow_pausing"
WORKFLOW_PAUSING_TASK_ACTIVE_ITEMS_INCOMPLETE = "workflow_pausing_task_active_items_incomplete"
WORKFLOW_PAUSING_TASK_DORMANT_ITEMS_INCOMPLETE = "workflow_pausing_task_dormant_items_incomplete"
WORKFLOW_PAUSING_WORKFLOW_ACTIVE = "workflow_pausing_workflow_active"
WORKFLOW_PAUSING_WORKFLOW_DORMANT = "workflow_pausing_workflow_dormant"
WORKFLOW_PAUSED = "workflow_paused"
WORKFLOW_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE = "workflow_paused_task_active_items_incomplete"
WORKFLOW_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE = "workflow_paused_task_dormant_items_incomplete"
WORKFLOW_PAUSED_WORKFLOW_ACTIVE = "workflow_paused_workflow_active"
WORKFLOW_PAUSED_WORKFLOW_DORMANT = "workflow_paused_workflow_dormant"
WORKFLOW_RESUMING = "workflow_resuming"
WORKFLOW_RESUMING_WORKFLOW_COMPLETED = "workflow_resuming_workflow_completed"
WORKFLOW_SUCCEEDED = "workflow_succeeded"
WORKFLOW_FAILED = "workflow_failed"
WORKFLOW_CANCELING = "workflow_canceling"
WORKFLOW_CANCELING_TASK_ACTIVE_ITEMS_INCOMPLETE = "workflow_canceling_task_active_items_incomplete"
WORKFLOW_CANCELING_TASK_DORMANT_ITEMS_INCOMPLETE = (
    "workflow_canceling_task_dormant_items_incomplete"
)
WORKFLOW_CANCELING_WORKFLOW_ACTIVE = "workflow_canceling_workflow_active"
WORKFLOW_CANCELING_WORKFLOW_DORMANT = "workflow_canceling_workflow_dormant"
WORKFLOW_CANCELED = "workflow_canceled"
WORKFLOW_CANCELED_TASK_ACTIVE_ITEMS_INCOMPLETE = "workflow_canceled_task_active_items_incomplete"
WORKFLOW_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE = "workflow_canceled_task_dormant_items_incomplete"
WORKFLOW_CANCELED_WORKFLOW_ACTIVE = "workflow_canceled_workflow_active"
WORKFLOW_CANCELED_WORKFLOW_DORMANT = "workflow_canceled_workflow_dormant"

WORKFLOW_EXECUTION_EVENTS = [
    WORKFLOW_REQUESTED,
    WORKFLOW_SCHEDULED,
    WORKFLOW_DELAYED,
    WORKFLOW_RUNNING,
    WORKFLOW_RUNNING_WORKFLOW_COMPLETED,
    WORKFLOW_PAUSING,
    WORKFLOW_PAUSING_TASK_ACTIVE_ITEMS_INCOMPLETE,
    WORKFLOW_PAUSING_TASK_DORMANT_ITEMS_INCOMPLETE,
    WORKFLOW_PAUSING_WORKFLOW_ACTIVE,
    WORKFLOW_PAUSING_WORKFLOW_DORMANT,
    WORKFLOW_PAUSED,
    WORKFLOW_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    WORKFLOW_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE,
    WORKFLOW_PAUSED_WORKFLOW_ACTIVE,
    WORKFLOW_PAUSED_WORKFLOW_DORMANT,
    WORKFLOW_RESUMING,
    WORKFLOW_RESUMING_WORKFLOW_COMPLETED,
    WORKFLOW_SUCCEEDED,
    WORKFLOW_FAILED,
    WORKFLOW_CANCELING,
    WORKFLOW_CANCELING_TASK_ACTIVE_ITEMS_INCOMPLETE,
    WORKFLOW_CANCELING_TASK_DORMANT_ITEMS_INCOMPLETE,
    WORKFLOW_CANCELING_WORKFLOW_ACTIVE,
    WORKFLOW_CANCELING_WORKFLOW_DORMANT,
    WORKFLOW_CANCELED,
    WORKFLOW_CANCELED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    WORKFLOW_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE,
    WORKFLOW_CANCELED_WORKFLOW_ACTIVE,
    WORKFLOW_CANCELED_WORKFLOW_DORMANT,
]


# Events for task execution.
TASK_REQUESTED = "task_requested"
TASK_SCHEDULED = "task_scheduled"
TASK_DELAYED = "task_delayed"
TASK_RUNNING = "task_running"
TASK_PENDING = "task_pending"
TASK_PENDING_WORKFLOW_ACTIVE = "task_pending_workflow_active"
TASK_PENDING_WORKFLOW_DORMANT = "task_pending_workflow_dormant"
TASK_PAUSING = "task_pausing"
TASK_PAUSED = "task_paused"
TASK_PAUSED_WORKFLOW_ACTIVE = "task_paused_workflow_active"
TASK_PAUSED_WORKFLOW_DORMANT = "task_paused_workflow_dormant"
TASK_RESUMING = "task_resuming"
TASK_SUCCEEDED = "task_succeeded"
TASK_SUCCEEDED_WORKFLOW_ACTIVE_INCOMPLETE = "task_succeeded_workflow_active_incomplete"
TASK_SUCCEEDED_WORKFLOW_ACTIVE_COMPLETED = "task_succeeded_workflow_active_completed"
TASK_SUCCEEDED_WORKFLOW_ACTIVE_PAUSED = "task_succeeded_workflow_active_paused"
TASK_SUCCEEDED_WORKFLOW_ACTIVE_CANCELED = "task_succeeded_workflow_active_canceled"
TASK_SUCCEEDED_WORKFLOW_DORMANT_INCOMPLETE = "task_succeeded_workflow_dormant_incomplete"
TASK_SUCCEEDED_WORKFLOW_DORMANT_COMPLETED = "task_succeeded_workflow_dormant_completed"
TASK_SUCCEEDED_WORKFLOW_DORMANT_PAUSED = "task_succeeded_workflow_dormant_paused"
TASK_SUCCEEDED_WORKFLOW_DORMANT_CANCELED = "task_succeeded_workflow_dormant_canceled"
TASK_FAILED = "task_failed"
TASK_FAILED_WORKFLOW_ACTIVE = "task_failed_workflow_active"
TASK_FAILED_WORKFLOW_DORMANT = "task_failed_workflow_dormant"
TASK_REMEDIATED = "task_remediated"
TASK_REMEDIATED_WORKFLOW_ACTIVE_INCOMPLETE = "task_remediated_workflow_active_incomplete"
TASK_REMEDIATED_WORKFLOW_ACTIVE_COMPLETED = "task_remediated_workflow_active_completed"
TASK_REMEDIATED_WORKFLOW_ACTIVE_PAUSED = "task_remediated_workflow_active_paused"
TASK_REMEDIATED_WORKFLOW_ACTIVE_CANCELED = "task_remediated_workflow_active_canceled"
TASK_REMEDIATED_WORKFLOW_DORMANT_INCOMPLETE = "task_remediated_workflow_dormant_incomplete"
TASK_REMEDIATED_WORKFLOW_DORMANT_COMPLETED = "task_remediated_workflow_dormant_completed"
TASK_REMEDIATED_WORKFLOW_DORMANT_PAUSED = "task_remediated_workflow_dormant_paused"
TASK_REMEDIATED_WORKFLOW_DORMANT_CANCELED = "task_remediated_workflow_dormant_canceled"
TASK_CANCELING = "task_canceling"
TASK_CANCELED = "task_canceled"
TASK_CANCELED_WORKFLOW_ACTIVE = "task_canceled_workflow_active"
TASK_CANCELED_WORKFLOW_DORMANT = "task_canceled_workflow_dormant"
TASK_RETRYING = "task_retrying"
TASK_RETRYING_WORKFLOW_ACTIVE = "task_retrying_workflow_active"
TASK_RETRYING_WORKFLOW_DORMANT = "task_retrying_workflow_dormant"

# Task events that require additional workflow context for processing.
# These are events related to the tasks being complete or inactive.
TASK_CONDITIONAL_EVENTS = [
    TASK_PENDING,
    TASK_PAUSED,
    TASK_SUCCEEDED,
    TASK_FAILED,
    TASK_REMEDIATED,
    TASK_CANCELED,
    TASK_RETRYING,
]

TASK_EXECUTION_EVENTS = [
    TASK_REQUESTED,
    TASK_SCHEDULED,
    TASK_DELAYED,
    TASK_RUNNING,
    TASK_PENDING,
    TASK_PENDING_WORKFLOW_ACTIVE,
    TASK_PENDING_WORKFLOW_DORMANT,
    TASK_PAUSING,
    TASK_PAUSED,
    TASK_PAUSED_WORKFLOW_ACTIVE,
    TASK_PAUSED_WORKFLOW_DORMANT,
    TASK_RESUMING,
    TASK_SUCCEEDED,
    TASK_SUCCEEDED_WORKFLOW_ACTIVE_INCOMPLETE,
    TASK_SUCCEEDED_WORKFLOW_ACTIVE_COMPLETED,
    TASK_SUCCEEDED_WORKFLOW_ACTIVE_PAUSED,
    TASK_SUCCEEDED_WORKFLOW_ACTIVE_CANCELED,
    TASK_SUCCEEDED_WORKFLOW_DORMANT_INCOMPLETE,
    TASK_SUCCEEDED_WORKFLOW_DORMANT_COMPLETED,
    TASK_SUCCEEDED_WORKFLOW_DORMANT_PAUSED,
    TASK_SUCCEEDED_WORKFLOW_DORMANT_CANCELED,
    TASK_FAILED,
    TASK_FAILED_WORKFLOW_ACTIVE,
    TASK_FAILED_WORKFLOW_DORMANT,
    TASK_REMEDIATED,
    TASK_REMEDIATED_WORKFLOW_ACTIVE_INCOMPLETE,
    TASK_REMEDIATED_WORKFLOW_ACTIVE_COMPLETED,
    TASK_REMEDIATED_WORKFLOW_ACTIVE_PAUSED,
    TASK_REMEDIATED_WORKFLOW_ACTIVE_CANCELED,
    TASK_REMEDIATED_WORKFLOW_DORMANT_INCOMPLETE,
    TASK_REMEDIATED_WORKFLOW_DORMANT_COMPLETED,
    TASK_REMEDIATED_WORKFLOW_DORMANT_PAUSED,
    TASK_REMEDIATED_WORKFLOW_DORMANT_CANCELED,
    TASK_CANCELING,
    TASK_CANCELED,
    TASK_CANCELED_WORKFLOW_ACTIVE,
    TASK_CANCELED_WORKFLOW_DORMANT,
    TASK_RETRYING,
    TASK_RETRYING_WORKFLOW_ACTIVE,
    TASK_RETRYING_WORKFLOW_DORMANT,
]


# Events for action executions.
ACTION_REQUESTED = "action_requested"
ACTION_SCHEDULED = "action_scheduled"
ACTION_DELAYED = "action_delayed"
ACTION_RUNNING = "action_running"
ACTION_PENDING = "action_pending"
ACTION_PENDING_TASK_ACTIVE_ITEMS_INCOMPLETE = "action_pending_task_active_items_incomplete"
ACTION_PENDING_TASK_DORMANT_ITEMS_INCOMPLETE = "action_pending_task_dormant_items_incomplete"
ACTION_PENDING_TASK_DORMANT_ITEMS_COMPLETED = "action_pending_task_dormant_items_completed"
ACTION_PENDING_TASK_DORMANT_ITEMS_PAUSED = "action_pending_task_dormant_items_paused"
ACTION_PENDING_TASK_DORMANT_ITEMS_CANCELED = "action_pending_task_dormant_items_canceled"
ACTION_PENDING_TASK_DORMANT_ITEMS_FAILED = "action_pending_task_dormant_items_failed"
ACTION_PAUSING = "action_pausing"
ACTION_PAUSED = "action_paused"
ACTION_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE = "action_paused_task_active_items_incomplete"
ACTION_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE = "action_paused_task_dormant_items_incomplete"
ACTION_PAUSED_TASK_DORMANT_ITEMS_COMPLETED = "action_paused_task_dormant_items_completed"
ACTION_PAUSED_TASK_DORMANT_ITEMS_PAUSED = "action_paused_task_dormant_items_paused"
ACTION_PAUSED_TASK_DORMANT_ITEMS_CANCELED = "action_paused_task_dormant_items_canceled"
ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED = "action_paused_task_dormant_items_failed"
ACTION_RESUMING = "action_resuming"
ACTION_SUCCEEDED = "action_succeeded"
ACTION_SUCCEEDED_TASK_ACTIVE_ITEMS_INCOMPLETE = "action_succeeded_task_active_items_incomplete"
ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_INCOMPLETE = "action_succeeded_task_dormant_items_incomplete"
ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_COMPLETED = "action_succeeded_task_dormant_items_completed"
ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_PAUSED = "action_succeeded_task_dormant_items_paused"
ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_CANCELED = "action_succeeded_task_dormant_items_canceled"
ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED = "action_succeeded_task_dormant_items_failed"
ACTION_FAILED = "action_failed"
ACTION_FAILED_TASK_ACTIVE_ITEMS_INCOMPLETE = "action_failed_task_active_items_incomplete"
ACTION_FAILED_TASK_DORMANT_ITEMS_INCOMPLETE = "action_failed_task_dormant_items_incomplete"
ACTION_FAILED_TASK_DORMANT_ITEMS_COMPLETED = "action_failed_task_dormant_items_completed"
ACTION_FAILED_TASK_DORMANT_ITEMS_PAUSED = "action_failed_task_dormant_items_paused"
ACTION_FAILED_TASK_DORMANT_ITEMS_CANCELED = "action_failed_task_dormant_items_canceled"
ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED = "action_failed_task_dormant_items_failed"
ACTION_EXPIRED = "action_timeout"
ACTION_EXPIRED_TASK_ACTIVE_ITEMS_INCOMPLETE = "action_timeout_task_active_items_incomplete"
ACTION_EXPIRED_TASK_DORMANT_ITEMS_INCOMPLETE = "action_timeout_task_dormant_items_incomplete"
ACTION_EXPIRED_TASK_DORMANT_ITEMS_COMPLETED = "action_timeout_task_dormant_items_completed"
ACTION_EXPIRED_TASK_DORMANT_ITEMS_PAUSED = "action_timeout_task_dormant_items_paused"
ACTION_EXPIRED_TASK_DORMANT_ITEMS_CANCELED = "action_timeout_task_dormant_items_canceled"
ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED = "action_timeout_task_dormant_items_failed"
ACTION_ABANDONED = "action_abandoned"
ACTION_ABANDONED_TASK_ACTIVE_ITEMS_INCOMPLETE = "action_abandoned_task_active_items_incomplete"
ACTION_ABANDONED_TASK_DORMANT_ITEMS_INCOMPLETE = "action_abandoned_task_dormant_items_incomplete"
ACTION_ABANDONED_TASK_DORMANT_ITEMS_COMPLETED = "action_abandoned_task_dormant_items_completed"
ACTION_ABANDONED_TASK_DORMANT_ITEMS_PAUSED = "action_abandoned_task_dormant_items_paused"
ACTION_ABANDONED_TASK_DORMANT_ITEMS_CANCELED = "action_abandoned_task_dormant_items_canceled"
ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED = "action_abandoned_task_dormant_items_failed"
ACTION_CANCELING = "action_canceling"
ACTION_CANCELED = "action_canceled"
ACTION_CANCELED_TASK_ACTIVE_ITEMS_INCOMPLETE = "action_canceled_task_active_items_incomplete"
ACTION_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE = "action_canceled_task_dormant_items_incomplete"
ACTION_CANCELED_TASK_DORMANT_ITEMS_COMPLETED = "action_canceled_task_dormant_items_completed"
ACTION_CANCELED_TASK_DORMANT_ITEMS_PAUSED = "action_canceled_task_dormant_items_paused"
ACTION_CANCELED_TASK_DORMANT_ITEMS_CANCELED = "action_canceled_task_dormant_items_canceled"
ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED = "action_canceled_task_dormant_items_failed"

ACTION_EXECUTION_EVENTS = [
    ACTION_REQUESTED,
    ACTION_SCHEDULED,
    ACTION_DELAYED,
    ACTION_RUNNING,
    ACTION_PENDING,
    ACTION_PENDING_TASK_ACTIVE_ITEMS_INCOMPLETE,
    ACTION_PENDING_TASK_DORMANT_ITEMS_INCOMPLETE,
    ACTION_PENDING_TASK_DORMANT_ITEMS_COMPLETED,
    ACTION_PENDING_TASK_DORMANT_ITEMS_PAUSED,
    ACTION_PENDING_TASK_DORMANT_ITEMS_CANCELED,
    ACTION_PENDING_TASK_DORMANT_ITEMS_FAILED,
    ACTION_PAUSING,
    ACTION_PAUSED,
    ACTION_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    ACTION_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE,
    ACTION_PAUSED_TASK_DORMANT_ITEMS_COMPLETED,
    ACTION_PAUSED_TASK_DORMANT_ITEMS_PAUSED,
    ACTION_PAUSED_TASK_DORMANT_ITEMS_CANCELED,
    ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED,
    ACTION_RESUMING,
    ACTION_SUCCEEDED,
    ACTION_SUCCEEDED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_INCOMPLETE,
    ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_COMPLETED,
    ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_PAUSED,
    ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_CANCELED,
    ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED,
    ACTION_FAILED,
    ACTION_FAILED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    ACTION_FAILED_TASK_DORMANT_ITEMS_INCOMPLETE,
    ACTION_FAILED_TASK_DORMANT_ITEMS_COMPLETED,
    ACTION_FAILED_TASK_DORMANT_ITEMS_PAUSED,
    ACTION_FAILED_TASK_DORMANT_ITEMS_CANCELED,
    ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED,
    ACTION_EXPIRED,
    ACTION_EXPIRED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    ACTION_EXPIRED_TASK_DORMANT_ITEMS_INCOMPLETE,
    ACTION_EXPIRED_TASK_DORMANT_ITEMS_COMPLETED,
    ACTION_EXPIRED_TASK_DORMANT_ITEMS_PAUSED,
    ACTION_EXPIRED_TASK_DORMANT_ITEMS_CANCELED,
    ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED,
    ACTION_ABANDONED,
    ACTION_ABANDONED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    ACTION_ABANDONED_TASK_DORMANT_ITEMS_INCOMPLETE,
    ACTION_ABANDONED_TASK_DORMANT_ITEMS_COMPLETED,
    ACTION_ABANDONED_TASK_DORMANT_ITEMS_PAUSED,
    ACTION_ABANDONED_TASK_DORMANT_ITEMS_CANCELED,
    ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED,
    ACTION_CANCELING,
    ACTION_CANCELED,
    ACTION_CANCELED_TASK_ACTIVE_ITEMS_INCOMPLETE,
    ACTION_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE,
    ACTION_CANCELED_TASK_DORMANT_ITEMS_COMPLETED,
    ACTION_CANCELED_TASK_DORMANT_ITEMS_PAUSED,
    ACTION_CANCELED_TASK_DORMANT_ITEMS_CANCELED,
    ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED,
]


# Events for special workflow engine operations.
TASK_CONTINUE_REQUESTED = "task_continue_requested"
TASK_FAIL_REQUESTED = "task_fail_requested"
TASK_NOOP_REQUESTED = "task_noop_requested"
TASK_RETRY_REQUESTED = "task_retry_requested"

ENGINE_OPERATION_EVENTS = [
    TASK_CONTINUE_REQUESTED,
    TASK_FAIL_REQUESTED,
    TASK_NOOP_REQUESTED,
    TASK_RETRY_REQUESTED,
]


class ExecutionEvent(object):
    def __init__(self, name, status, result=None, context=None):
        if not statuses.is_valid(status):
            raise exc.InvalidStatus(status)

        self.name = name
        self.status = status
        self.result = result
        self.context = context


class WorkflowExecutionEvent(ExecutionEvent):
    def __init__(self, status):
        super(WorkflowExecutionEvent, self).__init__("workflow_%s" % status, status)


class TaskExecutionEvent(ExecutionEvent):
    def __init__(self, task_id, route, status):
        super(TaskExecutionEvent, self).__init__("task_%s" % status, status)
        self.task_id = task_id
        self.route = route


class ActionExecutionEvent(ExecutionEvent):
    def __init__(self, status, result=None, context=None):
        super(ActionExecutionEvent, self).__init__(
            "action_%s" % status, status, result=result, context=context
        )


class TaskItemActionExecutionEvent(ActionExecutionEvent):
    def __init__(self, item_id, status, result=None, accumulated_result=None):
        super(TaskItemActionExecutionEvent, self).__init__(status, result=result)
        self.item_id = item_id
        self.accumulated_result = accumulated_result


class EngineOperationEvent(ExecutionEvent):
    pass


class TaskContinueEvent(EngineOperationEvent):
    def __init__(self):
        self.name = TASK_CONTINUE_REQUESTED
        self.status = statuses.SUCCEEDED
        self.result = None
        self.context = None


class TaskRetryEvent(EngineOperationEvent):
    def __init__(self):
        self.name = TASK_RETRY_REQUESTED
        self.status = statuses.RETRYING
        self.result = None
        self.context = None


class TaskNoopEvent(EngineOperationEvent):
    def __init__(self):
        self.name = TASK_NOOP_REQUESTED
        self.status = statuses.SUCCEEDED
        self.result = None
        self.context = None


class TaskFailEvent(EngineOperationEvent):
    def __init__(self):
        self.name = TASK_FAIL_REQUESTED
        self.status = statuses.FAILED
        self.result = None
        self.context = None


ENGINE_EVENT_MAP = {
    "continue": TaskContinueEvent,
    "fail": TaskFailEvent,
    "noop": TaskNoopEvent,
    "retry": TaskRetryEvent,
}



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from orquesta.composers import base as comp_base
from orquesta import graphing
from orquesta.specs import mock as mock_specs


LOG = logging.getLogger(__name__)


class WorkflowComposer(comp_base.WorkflowComposer):
    wf_spec_type = mock_specs.WorkflowSpec

    @classmethod
    def compose(cls, spec):
        if not cls.wf_spec_type:
            raise TypeError("Undefined spec type for composer.")

        if not isinstance(spec, cls.wf_spec_type):
            raise TypeError('Unsupported spec type "%s".' % str(type(spec)))

        wf_graph = graphing.WorkflowGraph()

        return wf_graph



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.utils import parameters as args_util


class InlineParametersTest(unittest.TestCase):
    def test_parse_basic(self):
        tests = [
            ("x=null", {"x": None}),
            ("x=123", {"x": 123}),
            ("x=-123", {"x": -123}),
            ("x=0.123", {"x": 0.123}),
            ("x=-0.123", {"x": -0.123}),
            ("x=true", {"x": True}),
            ("x=false", {"x": False}),
            ('x="abc"', {"x": "abc"}),
            ("x='abc'", {"x": "abc"}),
        ]

        for s, d in tests:
            self.assertDictEqual(args_util.parse_inline_params(s)[0], d)

    def test_parse_dictionary(self):
        tests = [("x='{\"a\": 1}'", {"x": {"a": 1}})]

        for s, d in tests:
            self.assertDictEqual(args_util.parse_inline_params(s)[0], d)

    def test_parse_expression(self):
        tests = [("x=<% $.abc %>", {"x": "<% $.abc %>"}), ("x={{ _.abc }}", {"x": "{{ _.abc }}"})]

        for s, d in tests:
            self.assertDictEqual(args_util.parse_inline_params(s)[0], d)

    def test_parse_multiple(self):
        tests = [
            ('x="abc" y="def"', [{"x": "abc"}, {"y": "def"}]),
            ('y="def" x="abc"', [{"y": "def"}, {"x": "abc"}]),
            ('x="abc", y="def"', [{"x": "abc"}, {"y": "def"}]),
            ('y="def", x="abc"', [{"y": "def"}, {"x": "abc"}]),
            ('x="abc"; y="def"', [{"x": "abc"}, {"y": "def"}]),
            ('y="def"; x="abc"', [{"y": "def"}, {"x": "abc"}]),
        ]

        for s, d in tests:
            self.assertListEqual(args_util.parse_inline_params(s), d)

    def test_parse_combination(self):
        s = 'i=123 j="abc" k=true x=<% $.abc %> y={{ _.abc }} z=\'{"a": 1}\''

        d = [
            {"i": 123},
            {"j": "abc"},
            {"k": True},
            {"x": "<% $.abc %>"},
            {"y": "{{ _.abc }}"},
            {"z": {"a": 1}},
        ]

        self.assertListEqual(args_util.parse_inline_params(s), d)

    def test_parse_empty_string(self):
        self.assertListEqual(args_util.parse_inline_params(str()), [])

    def test_parse_null_type(self):
        self.assertListEqual(args_util.parse_inline_params(None), [])

    def test_parse_bool_input(self):
        tests = [
            ("x=true", [{"x": True}]),
            ("y=True", [{"y": True}]),
            ("c=TRUE", [{"c": True}]),
            ("x=false", [{"x": False}]),
            ("y=False", [{"y": False}]),
            ("c=FALSE", [{"c": False}]),
        ]

        for s, d in tests:
            self.assertListEqual(args_util.parse_inline_params(s), d)

    def test_parse_string_in_quotes_input(self):
        tests = [
            ('x="true"', [{"x": "true"}]),
            ('y="True"', [{"y": "True"}]),
            ('c="TRUE"', [{"c": "TRUE"}]),
            ('d="123"', [{"d": "123"}]),
            ('e="abcde"', [{"e": "abcde"}]),
            ('f=""', [{"f": ""}]),
            ("x='false'", [{"x": "false"}]),
            ("y='False'", [{"y": "False"}]),
            ("c='FALSE'", [{"c": "FALSE"}]),
            ("d='123'", [{"d": "123"}]),
            ("e='abcde'", [{"e": "abcde"}]),
            ("f=''", [{"f": ""}]),
        ]

        for s, d in tests:
            self.assertListEqual(args_util.parse_inline_params(s), d)

    def test_parse_other_types(self):
        self.assertListEqual(args_util.parse_inline_params(123), [])
        self.assertListEqual(args_util.parse_inline_params(True), [])
        self.assertListEqual(args_util.parse_inline_params([1, 2, 3]), [])
        self.assertListEqual(args_util.parse_inline_params({"a": 123}), [])



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.specs.mistral import v2 as mistral_v2_specs
from orquesta.specs.mistral.v2 import tasks as task_models
from orquesta.specs.mistral.v2 import workflows as workflow_models


VERSION = mistral_v2_specs.VERSION
deserialize = workflow_models.deserialize
instantiate = workflow_models.instantiate
TaskDefaultsSpec = task_models.TaskDefaultsSpec
TaskSpec = task_models.TaskSpec
WorkflowSpec = workflow_models.WorkflowSpec

__all__ = [
    instantiate.__name__,
    deserialize.__name__,
    WorkflowSpec.__name__,
    TaskDefaultsSpec.__name__,
    TaskSpec.__name__,
]



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import statuses
from orquesta.tests.unit.conducting.mistral import base


class JoinWorkflowConductorTest(base.MistralWorkflowConductorTest):
    def test_join(self):
        wf_name = "join"

        expected_task_seq = ["task1", "task2", "task4", "task3", "task5", "task6", "task7"]

        self.assert_conducting_sequences(wf_name, expected_task_seq)

    def test_join_count(self):
        wf_name = "join-count"

        # Mock error at task6
        expected_task_seq = ["task1", "task2", "task4", "task6", "task3", "task5", "task8"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.RUNNING,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task8
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.RUNNING,
        )

        # Mock error at task7
        expected_task_seq = ["task1", "task2", "task4", "task6", "task3", "task5", "task7", "task8"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.RUNNING,  # task7
            statuses.SUCCEEDED,  # task8
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.RUNNING,
        )

    def test_join_count_with_branch_error(self):
        wf_name = "join-count"

        # Mock error at task6. The conductor runs breadth first
        # and so task3 and task5 has not started.
        expected_task_seq = ["task1", "task2", "task4", "task6"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task6
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.FAILED,
        )

        # Mock error at task7, note that task3 and task5 have
        # already satisfied the task8 join requirements.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            "task3",
            "task5",
            "task7",
            "task8",
            "task9",
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.FAILED,  # task7
            statuses.SUCCEEDED,  # task8
            statuses.SUCCEEDED,  # task9
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta import exceptions as exc
from orquesta import graphing
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base
from orquesta.utils import dictionary as dict_util
from orquesta.utils import jsonify as json_util


class WorkflowConductorTest(test_base.WorkflowConductorTest):
    def _prep_conductor(self, context=None, inputs=None, status=None):
        wf_def = """
        version: 1.0

        description: A basic sequential workflow.

        input:
          - a
          - b: False

        output:
          - data:
              a: <% ctx().a %>
              b: <% ctx().b %>
              c: <% ctx().c %>

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - c: 'xyz'
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3
          task3:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task4
          task4:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task5
          task5:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)

        kwargs = {
            "context": context if context is not None else None,
            "inputs": inputs if inputs is not None else None,
        }

        conductor = conducting.WorkflowConductor(spec, **kwargs)

        self.assertIsNone(conductor._graph)
        self.assertIsInstance(conductor.graph, graphing.WorkflowGraph)

        self.assertIsNone(conductor._workflow_state)
        self.assertIsInstance(conductor.workflow_state, conducting.WorkflowState)

        if status:
            self.assertEqual(conductor.workflow_state.status, statuses.UNSET)
            self.assertEqual(conductor.get_workflow_status(), statuses.UNSET)
            conductor.request_workflow_status(status)
            self.assertEqual(conductor.workflow_state.status, status)
            self.assertEqual(conductor.get_workflow_status(), status)
        else:
            self.assertEqual(conductor.workflow_state.status, statuses.UNSET)
            self.assertEqual(conductor.get_workflow_status(), statuses.UNSET)

        user_inputs = inputs or {}
        parent_context = context or {}

        self.assertDictEqual(conductor._inputs, user_inputs)
        self.assertDictEqual(conductor.get_workflow_input(), user_inputs)
        self.assertDictEqual(conductor._parent_ctx, parent_context)
        self.assertDictEqual(conductor.get_workflow_parent_context(), parent_context)

        default_inputs = {"a": None, "b": False}
        init_ctx_value = dict_util.merge_dicts(default_inputs, user_inputs, True)
        init_ctx_value = dict_util.merge_dicts(init_ctx_value, parent_context, True)
        self.assertDictEqual(conductor.get_workflow_initial_context(), init_ctx_value)

        return conductor

    def test_init(self):
        conductor = self._prep_conductor()

        # Serialize and check.
        data = conductor.serialize()

        expected_data = {
            "spec": conductor.spec.serialize(),
            "graph": conductor.graph.serialize(),
            "context": {},
            "input": {},
            "output": None,
            "errors": [],
            "log": [],
            "state": {
                "status": statuses.UNSET,
                "routes": [[]],
                "staged": [
                    {"id": "task1", "route": 0, "prev": {}, "ctxs": {"in": [0]}, "ready": True}
                ],
                "tasks": {},
                "sequence": [],
                "contexts": [{"a": None, "b": False}],
            },
        }

        self.assertDictEqual(data, expected_data)

        # Deserialize and check.
        conductor = conducting.WorkflowConductor.deserialize(data)

        self.assertIsInstance(conductor.spec, native_specs.WorkflowSpec)
        self.assertEqual(conductor.workflow_state.status, statuses.UNSET)
        self.assertEqual(conductor.get_workflow_status(), statuses.UNSET)
        self.assertIsInstance(conductor.graph, graphing.WorkflowGraph)
        self.assertEqual(len(conductor.graph._graph.node), 5)
        self.assertIsInstance(conductor.workflow_state, conducting.WorkflowState)

    def test_init_with_inputs(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs)

        # Serialize and check.
        data = conductor.serialize()

        expected_data = {
            "spec": conductor.spec.serialize(),
            "graph": conductor.graph.serialize(),
            "context": {},
            "input": inputs,
            "output": None,
            "errors": [],
            "log": [],
            "state": {
                "status": statuses.UNSET,
                "routes": [[]],
                "staged": [
                    {"id": "task1", "route": 0, "prev": {}, "ctxs": {"in": [0]}, "ready": True}
                ],
                "tasks": {},
                "sequence": [],
                "contexts": [inputs],
            },
        }

        self.assertDictEqual(data, expected_data)

        # Deserialize and check.
        conductor = conducting.WorkflowConductor.deserialize(data)

        self.assertIsInstance(conductor.spec, native_specs.WorkflowSpec)
        self.assertEqual(conductor.workflow_state.status, statuses.UNSET)
        self.assertEqual(conductor.get_workflow_status(), statuses.UNSET)
        self.assertIsInstance(conductor.graph, graphing.WorkflowGraph)
        self.assertEqual(len(conductor.graph._graph.node), 5)
        self.assertIsInstance(conductor.workflow_state, conducting.WorkflowState)

    def test_init_with_partial_inputs(self):
        inputs = {"a": 123}
        default_inputs = {"b": False}
        expected_initial_ctx = dict_util.merge_dicts(inputs, default_inputs, True)
        conductor = self._prep_conductor(inputs=inputs)

        # Serialize and check.
        data = conductor.serialize()

        expected_data = {
            "spec": conductor.spec.serialize(),
            "graph": conductor.graph.serialize(),
            "context": {},
            "input": inputs,
            "output": None,
            "errors": [],
            "log": [],
            "state": {
                "status": statuses.UNSET,
                "routes": [[]],
                "staged": [
                    {"id": "task1", "route": 0, "prev": {}, "ctxs": {"in": [0]}, "ready": True}
                ],
                "tasks": {},
                "sequence": [],
                "contexts": [expected_initial_ctx],
            },
        }

        self.assertDictEqual(data, expected_data)

        # Deserialize and check.
        conductor = conducting.WorkflowConductor.deserialize(data)

        self.assertIsInstance(conductor.spec, native_specs.WorkflowSpec)
        self.assertEqual(conductor.workflow_state.status, statuses.UNSET)
        self.assertEqual(conductor.get_workflow_status(), statuses.UNSET)
        self.assertIsInstance(conductor.graph, graphing.WorkflowGraph)
        self.assertEqual(len(conductor.graph._graph.node), 5)
        self.assertIsInstance(conductor.workflow_state, conducting.WorkflowState)

    def test_init_with_context(self):
        context = {"parent": {"ex_id": "12345"}}
        inputs = {"a": 123, "b": True}
        init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), json_util.deepcopy(context))

        conductor = self._prep_conductor(context=context, inputs=inputs)

        # Serialize and check.
        data = conductor.serialize()

        expected_data = {
            "spec": conductor.spec.serialize(),
            "graph": conductor.graph.serialize(),
            "context": context,
            "input": inputs,
            "output": None,
            "errors": [],
            "log": [],
            "state": {
                "status": statuses.UNSET,
                "routes": [[]],
                "staged": [
                    {"id": "task1", "route": 0, "prev": {}, "ctxs": {"in": [0]}, "ready": True}
                ],
                "tasks": {},
                "sequence": [],
                "contexts": [init_ctx],
            },
        }

        self.assertDictEqual(data, expected_data)

        # Deserialize and check.
        conductor = conducting.WorkflowConductor.deserialize(data)

        self.assertIsInstance(conductor.spec, native_specs.WorkflowSpec)
        self.assertEqual(conductor.workflow_state.status, statuses.UNSET)
        self.assertEqual(conductor.get_workflow_status(), statuses.UNSET)
        self.assertIsInstance(conductor.graph, graphing.WorkflowGraph)
        self.assertEqual(len(conductor.graph._graph.node), 5)
        self.assertIsInstance(conductor.workflow_state, conducting.WorkflowState)

    def test_serialization(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        # Mock task flows.
        for i in range(1, 6):
            status_changes = [statuses.RUNNING, statuses.SUCCEEDED]
            self.forward_task_statuses(conductor, "task" + str(i), status_changes)

        # Serialize and check.
        data = conductor.serialize()

        expected_data = {
            "spec": conductor.spec.serialize(),
            "graph": conductor.graph.serialize(),
            "state": conductor.workflow_state.serialize(),
            "context": conductor.get_workflow_parent_context(),
            "input": conductor.get_workflow_input(),
            "output": conductor.get_workflow_output(),
            "errors": conductor.errors,
            "log": conductor.log,
        }

        self.assertDictEqual(data, expected_data)

        # Deserialize and check.
        conductor = conducting.WorkflowConductor.deserialize(data)

        self.assertIsInstance(conductor.spec, native_specs.WorkflowSpec)
        self.assertIsInstance(conductor.graph, graphing.WorkflowGraph)
        self.assertEqual(len(conductor.graph._graph.node), 5)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertIsInstance(conductor.workflow_state, conducting.WorkflowState)
        self.assertEqual(len(conductor.workflow_state.tasks), 5)
        self.assertEqual(len(conductor.workflow_state.sequence), 5)

    def test_get_workflow_initial_context(self):
        conductor = self._prep_conductor()
        expected_init_ctx = {"a": None, "b": False}
        self.assertDictEqual(conductor.get_workflow_initial_context(), expected_init_ctx)

    def test_get_workflow_initial_context_with_inputs(self):
        inputs = {"a": 123, "b": True}
        expected_init_ctx = inputs
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)
        self.assertDictEqual(conductor.get_workflow_initial_context(), expected_init_ctx)

    def test_get_start_tasks(self):
        inputs = {"a": 123}
        expected_task_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)
        self.assert_next_task(conductor, "task1", expected_task_ctx)

    def test_get_start_tasks_when_graph_paused(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        conductor.request_workflow_status(statuses.PAUSING)
        self.assertListEqual(conductor.get_next_tasks(), [])

        conductor.request_workflow_status(statuses.PAUSED)
        self.assertListEqual(conductor.get_next_tasks(), [])

    def test_get_start_tasks_when_graph_canceled(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        conductor.request_workflow_status(statuses.CANCELING)
        self.assertListEqual(conductor.get_next_tasks(), [])

        conductor.request_workflow_status(statuses.CANCELED)
        self.assertListEqual(conductor.get_next_tasks(), [])

    def test_get_start_tasks_when_graph_abended(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        conductor.request_workflow_status(statuses.FAILED)
        self.assertListEqual(conductor.get_next_tasks(), [])

    def test_get_task(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_route = 0
        task_name = "task1"
        expected_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_ctx["__current_task"] = {"id": task_name, "route": task_route}
        expected_ctx["__state"] = conductor.workflow_state.serialize()
        task = conductor.get_task(task_name, task_route)
        self.assertEqual(task["id"], task_name)
        self.assertEqual(task["route"], task_route)
        self.assertDictEqual(task["ctx"], expected_ctx)

        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        task_name = "task2"
        expected_ctx = dict_util.merge_dicts(json_util.deepcopy(expected_ctx), {"c": "xyz"})
        expected_ctx["__current_task"] = {"id": task_name, "route": task_route}
        expected_ctx["__state"] = conductor.workflow_state.serialize()
        task = conductor.get_task(task_name, task_route)
        self.assertEqual(task["id"], task_name)
        self.assertEqual(task["route"], task_route)
        self.assertDictEqual(task["ctx"], expected_ctx)

    def test_get_next_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 5):
            task_name = "task" + str(i)
            next_task_name = "task" + str(i + 1)
            expected_task_ctx = {"a": 123, "b": False, "c": "xyz"}
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
            self.assert_next_task(conductor, next_task_name, expected_task_ctx)

    def test_get_next_tasks_repeatedly(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        next_task_name = "task2"
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})

        self.assert_next_task(conductor, task_name, expected_init_ctx)

        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])
        self.assert_next_task(conductor, has_next_task=False)
        self.assert_next_task(conductor, has_next_task=False)

        self.forward_task_statuses(conductor, task_name, [statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        self.forward_task_statuses(conductor, next_task_name, [statuses.RUNNING])
        self.assert_next_task(conductor, has_next_task=False)
        self.assert_next_task(conductor, has_next_task=False)

        self.forward_task_statuses(conductor, task_name, [statuses.SUCCEEDED])
        self.assert_next_task(conductor, has_next_task=False)
        self.assert_next_task(conductor, has_next_task=False)

    def test_get_next_tasks_when_this_task_paused(self):
        inputs = {"a": 123}
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        task_name = "task2"
        next_task_name = "task3"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.PAUSING])
        self.assert_next_task(conductor, has_next_task=False)
        self.forward_task_statuses(conductor, task_name, [statuses.PAUSED])
        self.assert_next_task(conductor, has_next_task=False)

        # After the previous task is paused, since there is no other tasks running,
        # the workflow is paused. The workflow needs to be resumed manually.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)
        conductor.request_workflow_status(statuses.RESUMING)
        self.assertEqual(conductor.get_workflow_status(), statuses.RESUMING)

        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

    def test_get_next_tasks_when_graph_paused(self):
        inputs = {"a": 123}
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        conductor.request_workflow_status(statuses.PAUSING)
        self.assert_next_task(conductor, has_next_task=False)

        conductor.request_workflow_status(statuses.PAUSED)
        self.assert_next_task(conductor, has_next_task=False)

        conductor.request_workflow_status(statuses.RESUMING)
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

    def test_get_next_tasks_when_this_task_canceled(self):
        inputs = {"a": 123}
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        task_name = "task2"
        next_task_name = "task3"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.CANCELING])
        self.assert_next_task(conductor, has_next_task=False)
        self.forward_task_statuses(conductor, task_name, [statuses.CANCELED])
        self.assert_next_task(conductor, has_next_task=False)

        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_get_next_tasks_when_graph_canceled(self):
        inputs = {"a": 123}
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        conductor.request_workflow_status(statuses.CANCELING)
        self.assert_next_task(conductor, has_next_task=False)

        conductor.request_workflow_status(statuses.CANCELED)
        self.assert_next_task(conductor, has_next_task=False)

    def test_get_next_tasks_when_this_task_abended(self):
        inputs = {"a": 123}
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        task_name = "task2"
        next_task_name = "task3"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.FAILED])
        self.assert_next_task(conductor, has_next_task=False)

        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

    def test_get_next_tasks_when_graph_abended(self):
        inputs = {"a": 123}
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        conductor.request_workflow_status(statuses.FAILED)
        self.assert_next_task(conductor, has_next_task=False)

    def test_get_task_initial_context(self):
        inputs = {"a": 123}
        expected_init_ctx = dict_util.merge_dicts(json_util.deepcopy(inputs), {"b": False})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_route = 0
        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        actual_task_ctx = conductor.get_task_initial_context(task_name, task_route)
        self.assertDictEqual(actual_task_ctx, expected_init_ctx)

        expected_task_out_ctx = {"c": "xyz"}
        expected_context_list = [expected_init_ctx, expected_task_out_ctx]
        self.assertListEqual(conductor.workflow_state.contexts, expected_context_list)

    def test_get_task_transition_contexts(self):
        inputs = {"a": 123, "b": True}
        expected_init_ctx = json_util.deepcopy(inputs)
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        # Get context for task2 that is staged by not yet running.
        task_route = 0
        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        next_task_in_ctx = expected_task_ctx
        expected_task_transition_ctx = {"%s__t0" % next_task_name: next_task_in_ctx}

        self.assertDictEqual(
            conductor.get_task_transition_contexts(task_name, task_route),
            expected_task_transition_ctx,
        )

        # Get context for task2 that is already running.
        task_name = "task2"
        next_task_name = "task3"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        expected_task_transition_ctx = {}

        self.assertDictEqual(
            conductor.get_task_transition_contexts(task_name, task_route),
            expected_task_transition_ctx,
        )

        # Get context for task3 that is not staged yet.
        self.assertDictEqual(conductor.get_task_transition_contexts(task_name, task_route), {})

        # Get transition context for task3 that has not yet run.
        self.assertRaises(
            exc.InvalidTaskStateEntry,
            conductor.get_task_transition_contexts,
            next_task_name,
            task_route,
        )

    def test_get_terminal_tasks_when_workflow_incomplete(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 5):
            task_name = "task" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assertListEqual(conductor.workflow_state.get_terminal_tasks(), [])

    def test_get_terminal_tasks_when_workflow_completed(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 6):
            task_name = "task" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        term_tasks = conductor.workflow_state.get_terminal_tasks()
        actual_term_tasks = [t["id"] for i, t in term_tasks]
        expected_term_tasks = ["task5"]
        self.assertListEqual(actual_term_tasks, expected_term_tasks)

    def test_get_terminal_tasks_when_workflow_failed(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 4):
            task_name = "task" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        task_name = "task4"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.FAILED])

        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

        term_tasks = conductor.workflow_state.get_terminal_tasks()
        actual_term_tasks = [t["id"] for i, t in term_tasks]
        expected_term_tasks = ["task4"]
        self.assertListEqual(actual_term_tasks, expected_term_tasks)

    def test_get_workflow_terminal_context_when_workflow_incomplete(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)
        self.assertRaises(exc.WorkflowContextError, conductor.get_workflow_terminal_context)

    def test_get_workflow_terminal_context_when_workflow_completed(self):
        inputs = {"a": 123, "b": True}
        expected_init_ctx = json_util.deepcopy(inputs)
        expected_term_ctx = json_util.deepcopy(expected_init_ctx)
        expected_term_ctx = dict_util.merge_dicts(expected_term_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 6):
            task_name = "task" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_terminal_context(), expected_term_ctx)

    def test_get_workflow_output_when_workflow_incomplete(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 5):
            task_name = "task" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assertIsNone(conductor.get_workflow_output())

    def test_get_workflow_output_when_workflow_failed(self):
        inputs = {"a": 123, "b": True}
        expected_init_ctx = json_util.deepcopy(inputs)
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 5):
            task_name = "task" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        task_name = "task5"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.FAILED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        expected_output = {"data": expected_task_ctx}
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_get_workflow_output_when_workflow_succeeded(self):
        inputs = {"a": 123, "b": True}
        expected_init_ctx = json_util.deepcopy(inputs)
        expected_task_ctx = json_util.deepcopy(expected_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"c": "xyz"})
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        for i in range(1, 6):
            task_name = "task" + str(i)
            self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        expected_output = {"data": expected_task_ctx}
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_set_workflow_canceling_when_no_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_set_workflow_canceled_when_no_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        conductor.request_workflow_status(statuses.CANCELED)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_set_workflow_canceling_when_has_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

    def test_set_workflow_canceled_when_has_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        conductor.request_workflow_status(statuses.CANCELED)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

    def test_set_workflow_pausing_when_no_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_set_workflow_paused_when_no_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        conductor.request_workflow_status(statuses.PAUSED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_set_workflow_pausing_when_has_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

    def test_set_workflow_paused_when_has_active_tasks(self):
        inputs = {"a": 123}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        conductor.request_workflow_status(statuses.PAUSED)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

    def test_append_log_entries(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        extra = {"x": 1234}
        conductor.log_entry("info", "The workflow is running as expected.", data=extra)
        conductor.log_entry("warn", "The task may be running a little bit slow.", task_id="task1")
        conductor.log_entry("error", "This is baloney.", task_id="task1")
        conductor.log_error(TypeError("Something is not right."), task_id="task1")
        conductor.log_errors([KeyError("task1"), ValueError("foobar")], task_id="task1")

        self.assertRaises(
            exc.WorkflowLogEntryError, conductor.log_entry, "foobar", "This is foobar."
        )

        expected_log_entries = [
            {"type": "info", "message": "The workflow is running as expected.", "data": extra},
            {
                "type": "warn",
                "message": "The task may be running a little bit slow.",
                "task_id": "task1",
            },
        ]

        expected_errors = [
            {"type": "error", "message": "This is baloney.", "task_id": "task1"},
            {"type": "error", "message": "TypeError: Something is not right.", "task_id": "task1"},
            {"type": "error", "message": "KeyError: 'task1'", "task_id": "task1"},
            {"type": "error", "message": "ValueError: foobar", "task_id": "task1"},
        ]

        self.assertListEqual(conductor.log, expected_log_entries)
        self.assertListEqual(conductor.errors, expected_errors)

        # Serialize and check.
        data = conductor.serialize()

        expected_data = {
            "spec": conductor.spec.serialize(),
            "graph": conductor.graph.serialize(),
            "context": {},
            "input": inputs,
            "output": None,
            "errors": expected_errors,
            "log": expected_log_entries,
            "state": {
                "status": statuses.RUNNING,
                "routes": [[]],
                "staged": [
                    {"id": "task1", "route": 0, "prev": {}, "ctxs": {"in": [0]}, "ready": True}
                ],
                "tasks": {},
                "sequence": [],
                "contexts": [inputs],
            },
        }

        self.assertDictEqual(data, expected_data)

        # Deserialize and check.
        conductor = conducting.WorkflowConductor.deserialize(data)

        self.assertIsInstance(conductor.spec, native_specs.WorkflowSpec)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)
        self.assertIsInstance(conductor.graph, graphing.WorkflowGraph)
        self.assertEqual(len(conductor.graph._graph.node), 5)
        self.assertIsInstance(conductor.workflow_state, conducting.WorkflowState)
        self.assertListEqual(conductor.log, expected_log_entries)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_append_duplicate_log_entries(self):
        inputs = {"a": 123, "b": True}
        conductor = self._prep_conductor(inputs=inputs, status=statuses.RUNNING)

        extra = {"x": 1234}
        task = {"task_id": "task1", "route": 0}
        conductor.log_entry("info", "The workflow is running as expected.", data=extra)
        conductor.log_entry("info", "The workflow is running as expected.", data=extra)
        conductor.log_entry("warn", "The task may be running a little bit slow.", **task)
        conductor.log_entry("warn", "The task may be running a little bit slow.", **task)
        conductor.log_entry("error", "This is baloney.", **task)
        conductor.log_entry("error", "This is baloney.", **task)
        conductor.log_error(TypeError("Something is not right."), **task)
        conductor.log_error(TypeError("Something is not right."), **task)
        conductor.log_errors([KeyError("task1"), ValueError("foobar")], **task)
        conductor.log_errors([KeyError("task1"), ValueError("foobar")], **task)

        expected_log_entries = [
            {"type": "info", "message": "The workflow is running as expected.", "data": extra},
            {
                "type": "warn",
                "message": "The task may be running a little bit slow.",
                "task_id": "task1",
                "route": 0,
            },
        ]

        expected_errors = [
            {"type": "error", "message": "This is baloney.", "task_id": "task1", "route": 0},
            {
                "type": "error",
                "message": "TypeError: Something is not right.",
                "task_id": "task1",
                "route": 0,
            },
            {"type": "error", "message": "KeyError: 'task1'", "task_id": "task1", "route": 0},
            {"type": "error", "message": "ValueError: foobar", "task_id": "task1", "route": 0},
        ]

        self.assertListEqual(conductor.log, expected_log_entries)
        self.assertListEqual(conductor.errors, expected_errors)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import unittest

import dateutil.tz

from orquesta.utils import date as date_util


class DateTest(unittest.TestCase):
    def test_date_valid(self):
        self.assertFalse(date_util.valid(123))
        self.assertFalse(date_util.valid(False))
        self.assertFalse(date_util.valid("abcde"))
        self.assertTrue(date_util.valid(datetime.datetime.utcnow()))
        self.assertTrue(date_util.valid(str(datetime.datetime.utcnow())))
        self.assertTrue(date_util.valid("2015-01-01 12:00:01.000000Z"))
        self.assertTrue(date_util.valid("2015-01-01 12:00:01.000000+04"))
        self.assertTrue(date_util.valid("2015-01-01 12:00:01.000000+0600"))
        self.assertTrue(date_util.valid("2015-01-01 12:00:01.000000+08:30"))

    def test_date_parse(self):
        offset = dateutil.tz.tzoffset(None, 3600)
        expected = datetime.datetime(2015, 1, 1, 12, 0, 1, tzinfo=offset)
        dt_str = "2015-01-01T12:00:01.000000+01:00"
        self.assertEqual(expected, date_util.parse(dt_str))

    def test_date_format(self):
        offset = dateutil.tz.tzoffset(None, 3600)
        dt = datetime.datetime(2015, 1, 1, 12, 0, 1, tzinfo=offset)
        expected = "2015-01-01T12:00:01.000000+01:00"
        self.assertEqual(expected, date_util.format(dt))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from orquesta.specs import base as spec_base


LOG = logging.getLogger(__name__)


def instantiate(definition):
    if len(definition.keys()) > 1:
        raise ValueError("Workflow definition contains more than one workflow.")

    wf_name, wf_spec = list(definition.items())[0]

    return WorkflowSpec(wf_spec, name=wf_name)


def deserialize(data):
    return WorkflowSpec.deserialize(data)


class WorkflowSpec(spec_base.Spec):
    _catalog = "mock"



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from orquesta.utils import jsonify as json_util


LOG = logging.getLogger(__name__)


def set_current_task(context, task):
    if context and not isinstance(context, dict):
        raise TypeError("The context is not type of dict.")

    if not task:
        raise ValueError("The task is not set.")

    if not isinstance(task, dict):
        raise TypeError("The task is not type of dict.")

    ctx = json_util.deepcopy(context) if context else dict()

    ctx["__current_task"] = {"id": task.get("id"), "route": task.get("route")}

    if "result" in task:
        ctx["__current_task"]["result"] = task.get("result")

    return ctx


def set_current_item(context, item):
    if context and not isinstance(context, dict):
        raise TypeError("The context is not type of dict.")

    ctx = json_util.deepcopy(context) if context else dict()
    ctx["__current_item"] = item

    return ctx



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.specs import base as spec_base
from orquesta.specs.mock import models as mock_models


VERSION = spec_base.Spec.get_version()
deserialize = mock_models.deserialize
instantiate = mock_models.instantiate
WorkflowSpec = mock_models.WorkflowSpec

__all__ = [instantiate.__name__, deserialize.__name__, WorkflowSpec.__name__]



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta import events
from orquesta import exceptions as exc
from orquesta import graphing
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorTaskStateTest(test_base.WorkflowConductorTest):
    def _add_tasks(self, wf_graph):
        for i in range(1, 6):
            task_name = "task" + str(i)
            wf_graph.add_task(task_name, name=task_name)

    def _add_transitions(self, wf_graph):
        wf_graph.add_transition("task1", "task2")
        wf_graph.add_transition("task1", "task5")
        wf_graph.add_transition("task2", "task3")
        wf_graph.add_transition("task3", "task4")
        wf_graph.add_transition("task4", "task2")

    def _prep_graph(self):
        wf_graph = graphing.WorkflowGraph()

        self._add_tasks(wf_graph)
        self._add_transitions(wf_graph)

        return wf_graph

    def _prep_conductor(self, context=None, inputs=None, status=None):
        wf_def = """
        version: 1.0

        description: A basic sequential workflow.

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task2, task5
          task2:
            action: core.noop
            next:
              - do: task3
          task3:
            action: core.noop
            next:
              - do: task4
          task4:
            action: core.noop
            next:
              - do: task2
          task5:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)

        kwargs = {
            "context": context if context is not None else None,
            "inputs": inputs if inputs is not None else None,
        }

        conductor = conducting.WorkflowConductor(spec, **kwargs)

        if status:
            conductor.request_workflow_status(status)

        return conductor

    def test_add_task_state(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        task_route = 0
        task_name = "task1"
        conductor.add_task_state(task_name, task_route)

        expected_task_state_idx = 0
        actual_task_state_idx = conductor._get_task_state_idx(task_name, task_route)
        self.assertEqual(actual_task_state_idx, expected_task_state_idx)

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "next": {},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

    def test_add_task_state_no_context(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        task_route = 0
        task_name = "task1"
        conductor.add_task_state(task_name, task_route)

        expected_task_state_idx = 0
        actual_task_state_idx = conductor._get_task_state_idx(task_name, task_route)
        self.assertEqual(actual_task_state_idx, expected_task_state_idx)

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "next": {},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

    def test_update_task_state(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        task_route = 0
        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        expected_task_state_idx = 0
        actual_task_state_idx = conductor._get_task_state_idx(task_name, task_route)
        self.assertEqual(actual_task_state_idx, expected_task_state_idx)

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "running",
            "next": {},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        self.forward_task_statuses(conductor, task_name, [statuses.SUCCEEDED], result="foobar")

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "succeeded",
            "next": {"task2__t0": True, "task5__t0": True},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

    def test_update_task_state_with_failed_event(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        task_route = 0
        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        expected_task_state_idx = 0
        actual_task_state_idx = conductor._get_task_state_idx(task_name, task_route)
        self.assertEqual(actual_task_state_idx, expected_task_state_idx)

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "running",
            "next": {},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        self.forward_task_statuses(
            conductor, task_name, [statuses.FAILED], result={"stdout": "boom!"}
        )

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "failed",
            "next": {"task2__t0": False, "task5__t0": False},
            "prev": {},
            "term": True,
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        expected_errors = [
            {
                "type": "error",
                "message": "Execution failed. See result for details.",
                "task_id": "task1",
                "result": {"stdout": "boom!"},
            }
        ]

        self.assertListEqual(conductor.errors, expected_errors)

    def test_update_task_state_for_not_ready_task(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        self.assertRaises(
            exc.InvalidTaskStateEntry,
            conductor.update_task_state,
            "task2",
            0,
            events.ActionExecutionEvent(statuses.RUNNING),
        )

    def test_update_task_state_for_nonexistent_task(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        self.assertRaises(
            exc.InvalidTask,
            conductor.update_task_state,
            "task999",
            0,
            events.ActionExecutionEvent(statuses.RUNNING),
        )

    def test_update_invalid_status_to_task_state_item(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        task_route = 0
        task_name = "task1"

        self.assertRaises(exc.InvalidStatus, events.ActionExecutionEvent, "foobar")

        self.assertRaises(TypeError, conductor.update_task_state, task_name, task_route, "foobar")

        self.forward_task_statuses(conductor, task_name, [statuses.REQUESTED])

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "requested",
            "next": {},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        # When transition is not valid, the task status is not changed. For the test case below,
        # the status change from requested to succeeded is not a valid transition.
        self.forward_task_statuses(conductor, task_name, [statuses.SUCCEEDED])

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "requested",
            "next": {},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

    def test_add_sequence_to_task_state(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)

        # Update progress of task1 to task flow.
        task_route = 0
        task_name = "task1"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        expected_task_state_idx = 0
        actual_task_state_idx = conductor._get_task_state_idx(task_name, task_route)
        self.assertEqual(actual_task_state_idx, expected_task_state_idx)

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "running",
            "next": {},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        self.forward_task_statuses(conductor, task_name, [statuses.SUCCEEDED], result="foobar")

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "succeeded",
            "next": {"task2__t0": True, "task5__t0": True},
            "prev": {},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        expected_sequence = [expected_task_state_entry]
        self.assertListEqual(conductor.workflow_state.sequence, expected_sequence)

        # Update progress of task2 to task flow.
        task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING])

        expected_task_state_idx = 1
        actual_task_state_idx = conductor._get_task_state_idx(task_name, task_route)
        self.assertEqual(actual_task_state_idx, expected_task_state_idx)

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "running",
            "next": {},
            "prev": {"task1__t0": 0},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        self.forward_task_statuses(conductor, task_name, [statuses.SUCCEEDED], result="foobar")

        expected_task_state_entry = {
            "id": task_name,
            "route": task_route,
            "ctxs": {"in": [0]},
            "status": "succeeded",
            "next": {"task3__t0": True},
            "prev": {"task1__t0": 0},
        }

        actual_task_state_entry = conductor.get_task_state_entry(task_name, task_route)
        self.assertDictEqual(actual_task_state_entry, expected_task_state_entry)

        expected_sequence.append(expected_task_state_entry)
        self.assertListEqual(conductor.workflow_state.sequence, expected_sequence)

    def test_add_cycle_to_task_state(self):
        conductor = self._prep_conductor(status=statuses.RUNNING)
        task_route = 0

        # Check that there's a cycle in the graph.
        self.assertFalse(conductor.graph.in_cycle("task1"))
        self.assertTrue(conductor.graph.in_cycle("task2"))
        self.assertTrue(conductor.graph.in_cycle("task3"))
        self.assertTrue(conductor.graph.in_cycle("task4"))

        # Add a cycle to the task flow.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        self.forward_task_statuses(conductor, "task4", [statuses.RUNNING, statuses.SUCCEEDED])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])

        # Check the reference pointer. Task2 points to the latest instance.
        self.assertEqual(conductor._get_task_state_idx("task1", task_route), 0)
        self.assertEqual(conductor._get_task_state_idx("task2", task_route), 4)
        self.assertEqual(conductor._get_task_state_idx("task3", task_route), 2)
        self.assertEqual(conductor._get_task_state_idx("task4", task_route), 3)

        # Check sequence.
        expected_sequence = [
            {
                "id": "task1",
                "route": task_route,
                "ctxs": {"in": [0]},
                "status": "succeeded",
                "next": {"task2__t0": True, "task5__t0": True},
                "prev": {},
            },
            {
                "id": "task2",
                "route": task_route,
                "ctxs": {"in": [0]},
                "status": "succeeded",
                "next": {"task3__t0": True},
                "prev": {"task1__t0": 0},
            },
            {
                "id": "task3",
                "route": task_route,
                "ctxs": {"in": [0]},
                "status": "succeeded",
                "next": {"task4__t0": True},
                "prev": {"task2__t0": 1},
            },
            {
                "id": "task4",
                "route": task_route,
                "ctxs": {"in": [0]},
                "status": "succeeded",
                "next": {"task2__t0": True},
                "prev": {"task3__t0": 2},
            },
            {
                "id": "task2",
                "route": task_route,
                "ctxs": {"in": [0]},
                "status": "running",
                "next": {},
                "prev": {"task4__t0": 3},
            },
        ]

        self.assertListEqual(conductor.workflow_state.sequence, expected_sequence)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorExtendedTest(test_base.WorkflowConductorTest):
    def test_self_looping(self):
        wf_def = """
        version: 1.0

        description: A basic workflow with cycle.

        vars:
          - loop: True

        tasks:
          task1:
            action: core.noop
            next:
              - do: task2
          task2:
            action: core.noop
            next:
              - when: <% ctx(loop) = true %>
                publish:
                  - loop: False
                do: task2
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Conduct task1 and check context and that there is no next tasks yet.
        task_route = 0
        task_name = "task1"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        expected_task_ctx = {"loop": True}
        expected_txsn_ctx = {"%s__t0" % next_task_name: expected_task_ctx}
        actual_txsn_ctx = conductor.get_task_transition_contexts(task_name, task_route)
        self.assertDictEqual(actual_txsn_ctx, expected_txsn_ctx)
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        # Conduct task2 and check next tasks and context.
        task_name = "task2"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        expected_task_ctx = {"loop": False}
        expected_txsn_ctx = {"%s__t0" % next_task_name: expected_task_ctx}
        actual_txsn_ctx = conductor.get_task_transition_contexts(task_name, task_route)
        self.assertDictEqual(actual_txsn_ctx, expected_txsn_ctx)
        self.assert_next_task(conductor, next_task_name, expected_task_ctx)

        # Conduct task2 and check next tasks and context.
        task_name = "task2"
        next_task_name = "task2"
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, has_next_task=False)

        # Check the list of tasks and make sure only the last entry for task2 is term=True.
        tasks = [task for task in conductor.workflow_state.sequence if task["id"] == "task2"]
        self.assertListEqual([task.get("term", False) for task in tasks], [False, True])

        # Check workflow status and context.
        expected_term_ctx = {"loop": False}
        self.assertDictEqual(conductor.get_workflow_terminal_context(), expected_term_ctx)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base
from orquesta.utils import dictionary as dict_util
from orquesta.utils import jsonify as json_util


class WorkflowConductorExtendedTest(test_base.WorkflowConductorTest):
    def test_join(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        tasks:
          # branch 1
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var1: 'xyz'
                do: task3

          # branch 2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var2: 123
                do: task3

          # adjoining branch
          task3:
            join: all
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var3: True
                do: task4
          task4:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Conduct task1 and task2 then check context and next tasks.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        self.assert_next_task(conductor, has_next_task=False)
        self.assertDictEqual(conductor.get_task_transition_contexts("task1", 0), {})

        # Succeed task1 and check context.
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        expected_task_ctx = {"var1": "xyz"}
        expected_txsn_ctx = {"task3__t0": expected_task_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task1", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, has_next_task=False)

        # Conduct task2 and check next tasks and context.
        self.forward_task_statuses(conductor, "task2", [statuses.SUCCEEDED])
        expected_task_ctx = {"var1": "xyz", "var2": 123}
        expected_txsn_ctx = {"task3__t0": expected_task_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task2", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, "task3", expected_task_ctx)

        # Conduct task3 and check merged context.
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        expected_task_init_ctx = expected_task_ctx
        self.assertDictEqual(conductor.get_task_initial_context("task3", 0), expected_task_init_ctx)
        expected_task_ctx = dict_util.merge_dicts(expected_task_ctx, {"var3": True})
        expected_txsn_ctx = {"task4__t0": expected_task_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task3", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, "task4", expected_task_ctx)

        # Conduct task4 and check final workflow status.
        self.forward_task_statuses(conductor, "task4", [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, has_next_task=False)

        # Check workflow status and context.
        expected_term_ctx = json_util.deepcopy(expected_task_ctx)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_terminal_context(), expected_term_ctx)

    def test_join_with_no_input_and_no_context_changes(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        tasks:
          # branch 1
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3

          # branch 2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3

          # adjoining branch
          task3:
            join: all
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task4
          task4:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Conduct task1 and task2 then check context and next tasks.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        self.assert_next_task(conductor, has_next_task=False)
        self.assertDictEqual(conductor.get_task_transition_contexts("task1", 0), {})

        # Succeed task1 and check context.
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        expected_init_ctx = dict()
        expected_txsn_ctx = {"task3__t0": expected_init_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task1", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, has_next_task=False)

        # Conduct task2 and check next tasks and context.
        self.forward_task_statuses(conductor, "task2", [statuses.SUCCEEDED])
        expected_txsn_ctx = {"task3__t0": expected_init_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task2", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, "task3", expected_init_ctx)

        # Conduct task3 and check merged context.
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        self.assertDictEqual(conductor.get_task_initial_context("task3", 0), expected_init_ctx)
        expected_txsn_ctx = {"task4__t0": expected_init_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task3", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, "task4", expected_init_ctx)

        # Conduct task4 and check final workflow status.
        self.forward_task_statuses(conductor, "task4", [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, has_next_task=False)

        # Check workflow status and context.
        expected_term_ctx = expected_init_ctx
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_terminal_context(), expected_term_ctx)

    def test_join_with_input_and_no_context_changes(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        input:
          - var1

        tasks:
          # branch 1
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3

          # branch 2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3

          # adjoining branch
          task3:
            join: all
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task4
          task4:
            action: core.noop
        """

        inputs = {"var1": "xyz"}
        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec, inputs=inputs)
        conductor.request_workflow_status(statuses.RUNNING)

        # Conduct task1 and task2 then check context and next tasks.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        self.assert_next_task(conductor, has_next_task=False)
        self.assertDictEqual(conductor.get_task_transition_contexts("task1", 0), {})

        # Succeed task1 and check context.
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        expected_task_ctx = json_util.deepcopy(inputs)
        expected_txsn_ctx = {"task3__t0": expected_task_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task1", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, has_next_task=False)

        # Conduct task2 and check next tasks and context.
        self.forward_task_statuses(conductor, "task2", [statuses.SUCCEEDED])
        expected_txsn_ctx = {"task3__t0": expected_task_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task2", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, "task3", expected_task_ctx)

        # Conduct task3 and check merged context.
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        self.assertDictEqual(conductor.get_task_initial_context("task3", 0), expected_task_ctx)
        expected_txsn_ctx = {"task4__t0": expected_task_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task3", 0), expected_txsn_ctx)
        self.assert_next_task(conductor, "task4", expected_task_ctx)

        # Conduct task4 and check final workflow status.
        self.forward_task_statuses(conductor, "task4", [statuses.RUNNING, statuses.SUCCEEDED])
        self.assert_next_task(conductor, has_next_task=False)

        # Check workflow status and context.
        expected_term_ctx = json_util.deepcopy(expected_task_ctx)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_terminal_context(), expected_term_ctx)

    def test_parallel(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - var1: abc
          - var2: 234

        tasks:
          # branch 1
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var1: 'xyz'
                do: task2
          task2:
            action: core.noop

          # branch 2
          task3:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var2: 123
                do: task4
          task4:
            action: core.noop

        output:
          - var1: <% ctx().var1 %>
          - var2: <% ctx().var2 %>
        """

        expected_output = {"var1": "xyz", "var2": 123}

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Conduct task1 and task3 and check context.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING])

        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        expected_t1_ctx = {"var1": "xyz", "var2": 234}
        expected_txsn_ctx = {"task2__t0": expected_t1_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task1", 0), expected_txsn_ctx)

        task_ids = ["task2"]
        task_ctxs = [expected_t1_ctx]
        task_routes = [0]
        self.assert_next_tasks(conductor, task_ids, task_ctxs, task_routes)

        self.forward_task_statuses(conductor, "task3", [statuses.SUCCEEDED])
        expected_t3_ctx = {"var1": "abc", "var2": 123}
        expected_txsn_ctx = {"task4__t0": expected_t3_ctx}
        self.assertDictEqual(conductor.get_task_transition_contexts("task3", 0), expected_txsn_ctx)

        task_ids = ["task2", "task4"]
        task_ctxs = [expected_t1_ctx, expected_t3_ctx]
        task_routes = [0] * 2
        self.assert_next_tasks(conductor, task_ids, task_ctxs, task_routes)

        # Conduct task2 and task4 and check context.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        self.forward_task_statuses(conductor, "task4", [statuses.RUNNING])

        self.forward_task_statuses(conductor, "task2", [statuses.SUCCEEDED])
        self.assertDictEqual(conductor.get_task_transition_contexts("task2", 0), {})
        self.assert_next_task(conductor, has_next_task=False)

        self.forward_task_statuses(conductor, "task4", [statuses.SUCCEEDED])
        self.assertDictEqual(conductor.get_task_transition_contexts("task4", 0), {})
        self.assert_next_task(conductor, has_next_task=False)

        # Check the list of terminal tasks.
        term_tasks = conductor.workflow_state.get_terminal_tasks()
        actual_term_tasks = [t["id"] for i, t in term_tasks]
        expected_term_tasks = ["task2", "task4"]
        self.assertListEqual(actual_term_tasks, expected_term_tasks)

        # Render workflow output and check workflow status and context.
        conductor.render_workflow_output()
        expected_term_ctx = {"var1": "xyz", "var2": 123}
        self.assertDictEqual(conductor.get_workflow_terminal_context(), expected_term_ctx)
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.tests.unit.specs import base as test_specs
from orquesta.utils import jsonify as json_util


class SpecTest(unittest.TestCase):
    def setUp(self):
        super(SpecTest, self).setUp()
        self.maxDiff = None

    def test_spec_serialization(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr1-1": "fubar",
            "attr1_2": "foosball",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
            "attr6": {"attr1": {"attr1": {"attr1": "wunderbar"}}},
            "attr7": [{"attr1": {"attr1": "wunderbar"}}, {"attr1": {"attr1": "wonderful"}}],
        }

        spec_obj_1 = test_specs.MockSpec(spec)

        expected_data = {
            "catalog": spec_obj_1.get_catalog(),
            "version": spec_obj_1.get_version(),
            "name": spec_obj_1.name,
            "spec": spec_obj_1.spec,
        }

        spec_obj_1_json = spec_obj_1.serialize()

        self.assertDictEqual(spec_obj_1_json, expected_data)

        spec_obj_2 = test_specs.MockSpec.deserialize(spec_obj_1_json)

        self.assertEqual(spec_obj_2.name, spec_obj_1.name)
        self.assertEqual(spec_obj_2.member, spec_obj_1.member)
        self.assertDictEqual(spec_obj_2.spec, spec_obj_1.spec)

    def test_spec_deserialization_errors(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr1-1": "fubar",
            "attr1_2": "foosball",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
            "attr6": {"attr1": {"attr1": {"attr1": "wunderbar"}}},
            "attr7": [{"attr1": {"attr1": "wunderbar"}}, {"attr1": {"attr1": "wonderful"}}],
        }

        spec_obj = test_specs.MockSpec(spec)

        spec_obj_json = spec_obj.serialize()

        # Test missing catalog information.
        test_json = json_util.deepcopy(spec_obj_json)
        test_json.pop("catalog")

        self.assertRaises(ValueError, test_specs.MockSpec.deserialize, test_json)

        # Test missing version information.
        test_json = json_util.deepcopy(spec_obj_json)
        test_json.pop("version")

        self.assertRaises(ValueError, test_specs.MockSpec.deserialize, test_json)

        # Test mismatch version information.
        test_json = json_util.deepcopy(spec_obj_json)
        test_json["version"] = "99.99"

        self.assertRaises(ValueError, test_specs.MockSpec.deserialize, test_json)

    def test_mapping_spec_serialization(self):
        spec = {"attr1": {"attr1": {"attr1": "wunderbar"}}}

        spec_obj_1 = test_specs.MockMappingSpec(spec, member=True)

        expected_data = {
            "catalog": spec_obj_1.get_catalog(),
            "version": spec_obj_1.get_version(),
            "member": spec_obj_1.member,
            "spec": spec_obj_1.spec,
        }

        spec_obj_1_json = spec_obj_1.serialize()

        self.assertDictEqual(spec_obj_1_json, expected_data)

        spec_obj_2 = test_specs.MockMappingSpec.deserialize(spec_obj_1_json)

        self.assertEqual(spec_obj_2.member, spec_obj_1.member)
        self.assertDictEqual(spec_obj_2.spec, spec_obj_1.spec)

    def test_sequence_spec_serialization(self):
        spec = [{"attr1": {"attr1": "wunderbar"}}, {"attr1": {"attr1": "wonderful"}}]

        spec_obj_1 = test_specs.MockSequenceSpec(spec, member=True)

        expected_data = {
            "catalog": spec_obj_1.get_catalog(),
            "version": spec_obj_1.get_version(),
            "member": spec_obj_1.member,
            "spec": spec_obj_1.spec,
        }

        spec_obj_1_json = spec_obj_1.serialize()

        self.assertDictEqual(spec_obj_1_json, expected_data)

        spec_obj_2 = test_specs.MockSequenceSpec.deserialize(spec_obj_1_json)

        self.assertEqual(spec_obj_2.member, spec_obj_1.member)
        self.assertListEqual(spec_obj_2.spec, spec_obj_1.spec)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta import exceptions as exc
from orquesta.expressions import base as expr_base


class YAQLFacadeEvaluationTest(unittest.TestCase):
    def test_basic_eval(self):
        data = {"foo": "bar"}

        self.assertEqual("bar", expr_base.evaluate("<% ctx().foo %>", data))
        self.assertEqual("foobar", expr_base.evaluate("foo<% ctx().foo %>", data))

    def test_basic_eval_undefined(self):
        expr = "<% ctx().foo %>"

        data = {}

        self.assertRaises(exc.ExpressionEvaluationException, expr_base.evaluate, expr, data)

    def test_nested_eval(self):
        expr = "<% ctx(ctx().foo) %>"

        data = {"foo": "bar", "bar": "foobar"}

        self.assertEqual("foobar", expr_base.evaluate(expr, data))

    def test_dict_eval(self):
        expr = "<% ctx().nested.foo %>"

        data = {"nested": {"foo": "bar"}}

        self.assertEqual("bar", expr_base.evaluate(expr, data))

    def test_multi_eval(self):
        expr = "<% ctx().foo %> and <% ctx().marco %>"

        data = {"foo": "bar", "marco": "polo"}

        self.assertEqual("bar and polo", expr_base.evaluate(expr, data))

    def test_eval_recursive(self):
        expr = "<% ctx().fee %>"

        data = {
            "fee": "<% ctx().fi %>",
            "fi": "<% ctx().fo %>",
            "fo": "<% ctx().fum %>",
            "fum": "fee-fi-fo-fum",
        }

        self.assertEqual("fee-fi-fo-fum", expr_base.evaluate(expr, data))

    def test_multi_eval_recursive(self):
        expr = "<% ctx().fee %> <% ctx().im %>"

        data = {
            "fee": "<% ctx().fi %>",
            "fi": "<% ctx().fo %>",
            "fo": "<% ctx().fum %>",
            "fum": "fee-fi-fo-fum!",
            "im": "<% ctx().hungry %>",
            "hungry": "i'm hungry!",
        }

        self.assertEqual("fee-fi-fo-fum! i'm hungry!", expr_base.evaluate(expr, data))

    def test_eval_list(self):
        expr = ["<% ctx().foo %>", "<% ctx().marco %>", "foo<% ctx().foo %>"]

        data = {"foo": "bar", "marco": "polo"}

        self.assertListEqual(["bar", "polo", "foobar"], expr_base.evaluate(expr, data))

    def test_eval_list_of_list(self):
        expr = [["<% ctx().foo %>", "<% ctx().marco %>"]]

        data = {"foo": "bar", "marco": "polo"}

        expected = [["bar", "polo"]]

        self.assertListEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_list_of_dict(self):
        expr = [{"foo": "<% ctx().bar %>", "<% ctx().marco %>": "polo"}]

        data = {"bar": "bar", "marco": "marco"}

        expected = [{"foo": "bar", "marco": "polo"}]

        self.assertListEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_dict(self):
        expr = {
            "foo": "<% ctx().bar %>",
            "<% ctx().marco %>": "polo",
            "foobar": "foo<% ctx().bar %>",
        }

        data = {"bar": "bar", "marco": "marco"}

        expected = {"foo": "bar", "marco": "polo", "foobar": "foobar"}

        self.assertDictEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_dict_of_dict(self):
        expr = {"nested": {"foo": "<% ctx().bar %>", "<% ctx().marco %>": "polo"}}

        data = {"bar": "bar", "marco": "marco"}

        expected = {"nested": {"foo": "bar", "marco": "polo"}}

        self.assertDictEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_dict_of_list(self):
        expr = {"nested": ["<% ctx().foo %>", "<% ctx().marco %>"]}

        data = {"foo": "bar", "marco": "polo"}

        expected = {"nested": ["bar", "polo"]}

        self.assertDictEqual(expected, expr_base.evaluate(expr, data))

    def test_type_preservation(self):
        data = {"k1": 101, "k2": 1.999, "k3": True, "k4": [1, 2], "k5": {"k": "v"}, "k6": None}

        self.assertEqual(data["k1"], expr_base.evaluate("<% ctx().k1 %>", data))

        self.assertEqual(data["k2"], expr_base.evaluate("<% ctx().k2 %>", data))

        self.assertTrue(expr_base.evaluate("<% ctx().k3 %>", data))

        self.assertListEqual(data["k4"], expr_base.evaluate("<% ctx().k4 %>", data))

        self.assertDictEqual(data["k5"], expr_base.evaluate("<% ctx().k5 %>", data))

        self.assertIsNone(expr_base.evaluate("<% ctx().k6 %>", data))

    def test_type_string_detection(self):
        expr = "<% ctx().foo %> -> <% ctx().bar %>"

        data = {"foo": 101, "bar": 201}

        self.assertEqual("101 -> 201", expr_base.evaluate(expr, data))

    def test_custom_function(self):
        expr = "<% json('{\"a\": 123}') %>"

        self.assertDictEqual({"a": 123}, expr_base.evaluate(expr))

    def test_custom_function_failure(self):
        expr = "<% json(int(123)) %>"

        self.assertRaises(exc.ExpressionEvaluationException, expr_base.evaluate, expr)

    def test_distinct_operator(self):
        test_cases = [
            {
                "expr": "<% ctx(val).distinct() %>",
                "input": {"val": [1, 2, 3, 1]},
                "expect": [1, 2, 3],
            },
            {
                "expr": "<% ctx(val).distinct() %>",
                "input": {"val": [{"a": 1}, {"b": 2}, {"a": 1}]},
                "expect": [{"a": 1}, {"b": 2}],
            },
            {
                "expr": "<% ctx(val).distinct($[1]) %>",
                "input": {"val": [["a", 1], ["b", 2], ["c", 1], ["a", 3]]},
                "expect": [["a", 1], ["b", 2], ["a", 3]],
            },
        ]
        for case in test_cases:
            self.assertEqual(case["expect"], expr_base.evaluate(case["expr"], case["input"]))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.composition.mistral import base


class TaskTransitionWorkflowComposerTest(base.MistralWorkflowComposerTest):
    def test_on_error(self):
        wf_name = "task-on-error"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                ],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                ],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_on_complete(self):
        wf_name = "task-on-complete"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}, {"id": "task4"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task4",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                ],
                [],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}, {"id": "task4"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task4",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                ],
                [],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_task_transitions_split(self):
        wf_name = "task-transitions-split"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2", "splits": ["task2"]}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                    {
                        "id": "task2",
                        "key": 1,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task2",
                        "key": 2,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2", "splits": ["task2"]}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-complete"),
                    },
                    {
                        "id": "task2",
                        "key": 1,
                        "criteria": self.compose_seq_expr("task1", condition="on-error"),
                    },
                    {
                        "id": "task2",
                        "key": 2,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)
